<?php
// $Id: authcache.helpers.inc,v 1.1 2009/03/08 00:56:55 jonahellison Exp $

/**
 * @file authcache.helpers.inc
 *
 * Helper functions for the Authcache module
 * (No Drupal hooks)
 */
 

/**
 * Should the current page be cached?
 */
function _authcache_is_cacheable() {
  global $user;

  // Do not cache POST, users with JS disabled, admin account,
  // Drupal messages, or PHP files (cron.php, update.php, etc)
  if(
    !empty($_POST)
    || ($user->uid && !isset($_COOKIE['has_js']))
    || $user->uid == 1
    || drupal_set_message() != NULL
    || substr($_SERVER['REQUEST_URI'], -4) == '.php'
    ) {
    return FALSE;
  }

  // Check if caching is enabled for user's role
  $cache_roles = variable_get('authcache_roles', array());
  
  // Anonymous
  if(!$user->uid && !in_array(DRUPAL_ANONYMOUS_RID, $cache_roles)) {
    return FALSE;
  }
  // Logged-in
  else if($user->uid) {
    unset($cache_roles[DRUPAL_ANONYMOUS_RID]);
    $extra_roles = array_diff(array_keys($user->roles), $cache_roles);
    
    if(!empty($extra_roles)) {
      return FALSE;
    }
  }

  $is_cached = FALSE;

  $alias = drupal_get_path_alias($_GET['q']);
  $path = drupal_get_normal_path($_GET['q']); // normalize path

  // Now check page caching settings, defined by the site admin
  $pagecaching = variable_get('authcache_pagecaching', array(array('option'=>0,'pages'=>AUTHCACHE_NOCACHE_DEFAULT,'roles'=>array(0,1))));
  foreach($pagecaching as $page_rules) {

    // Do caching page roles apply to current user?
    $extra_roles = array_diff(array_keys($user->roles), $page_rules['roles']);
    if(empty($extra_roles)) {
      switch($page_rules['option']) {
        case '0': // Cache every page except the listed pages.
        case '1': // Cache only the listed pages.
          $regexp = '/^('. preg_replace(array('/(\r\n?|\n)/', '/\\\\\*/', '/(^|\|)\\\\<front\\\\>($|\|)/'), array('|', '.*', '\1'. preg_quote(variable_get('site_frontpage', 'node'), '/') .'\2'), preg_quote($page_rules['pages'], '/')) .')$/';
          $is_cached = !($page_rules['option'] xor preg_match($regexp, $alias));
          break;
        case '2': // Cache pages for which the following PHP code returns TRUE
          $result = drupal_eval($page_rules['pages']);
          $is_cached = !empty($result);
          break;
        default:
          break;
      }
    }
  }
  
  return $is_cached;
}

/**
 * Save page to cache
 *
 * Called using PHP's register_shutdown_function()
 * This is better than an ob_start callback since global variables
 * are not deconstructed and the function is executed later
 */
function _authcache_shutdown_save_page() {
  global $user, $base_root, $is_page_authcache;
  
  // Final check, in case variable was modified
  if(!$is_page_authcache) {
    return;
  }
  
  // Don't cache pages with PHP errors (Drupal can't catch fatal errors)
  if(count(error_get_last())) {
    return;
  }
  
  // Only cache HTML and 200 OK status
  // Note that Druapl feeds don't overwrite content-type headers, so they will be cached
  if (_authcache_get_content_type() != 'text/html' && _authcache_get_http_status() != 200) {
    return;
  }

  // Don't append to CLI scripts
  if (empty($_SERVER['REQUEST_METHOD'])) {
    return;
  }
  
  // Make sure "Location" redirect isn't used
  foreach(headers_list() as $header) {
    if(strpos($header, "Location:") !== FALSE) {
      return;
    }
  }

  // Cache key, contructed from user role and URL
  $key = cache_authcache_key($user) . $base_root . request_uri();

  // Authcache info JSON
  $authcache_info = array(
    'page_render'         => timer_read('page'),  // Benchmark
    'page_queries'        => '-1',                // Database benchmark, if enabled
    'cache_render'        => '-1',                // Filled by cookie via JS on cache request
    'cache_uid'           => $user->uid,          // Required by JS for HTML updates
    'cache_time'          => time(),              // Required by JS for HTML updates
    //'cache_key' => $key,
  );
  
  // Database benchmarks
  if(variable_get('dev_query',0)) {
    global $queries;

    $time_query = 0;
    foreach($queries as $q) {
      $time_query += $q[1];
    }
    $time_query = round($time_query * 1000, 2); // Convert seconds to milliseconds
    $percent_query = round(($time_query / $authcache_info['page_render']) * 100);
    
    $authcache_info['page_queries'] = count($queries) . " queries @ {$time_query} ms ({$percent_query}%)";
  }
  else {
    unset($authcache_info['page_queries']);
  }

  // JSON to send via Ajax
  // The "q" key is need during Ajax phase
  $authcache_ajax = array(
    'q' => $_GET['q'],
  );
  
  // Invoke hook_authcache_info() operation to allow modules to modify info array
  _authcache_invoke_hook('authcache_info', $authcache_info);

  // Invoke hook_authcache() operation to allow modules to modify ajax array
  _authcache_invoke_hook('authcache_ajax', $authcache_ajax);

  // Get buffered HTML
  $buffer = ob_get_contents();
  ob_end_clean();
  
  // Don't cache empty/dead pages
  if(!$buffer) { 
    return;
  }
  
  $path = drupal_get_normal_path($_GET['q']); // normalize path
  
  // Only place JSON info for HTML pages
  $is_html = !($path == 'rss.xml' || preg_match('!/feed$!', $path));
  
  if($is_html) {
    $authcache_footer['info'] = $authcache_info;
    $authcache_footer['ajax'] = $authcache_ajax;
    $authcache_json =
    "\n<!-- Authcache Footer JSON -->\n" .
    "<script language=\"JavaScript\">\nvar authcacheFooter = ".
    json_encode($authcache_footer) .
    ";\n</script>\n";

    // Insert JSON before </body>, otherwise just append
    if(strripos($buffer,'</body>') !== FALSE) {
      $buffer = str_replace('</body>', $authcache_json.'</body>', $buffer);
    } else {
      $buffer .= $authcache_json;
    }

  }
  
  // Dump to browser, then max gzip & save to cache in background
  
  $output = $buffer;

  // Fast compression
  if (@strpos($_SERVER['HTTP_ACCEPT_ENCODING'], 'gzip') !== FALSE && variable_get('page_compression', TRUE) && function_exists('gzencode') && zlib_get_coding_type() == FALSE) {
    $output = gzencode($output, 1, FORCE_GZIP);
    header('Content-Encoding: gzip');
  }

  header("Content-Length: " . strlen($output));
  echo $output;
  flush();

  // Check for page compression
  if (variable_get('page_compression', TRUE) && function_exists('gzencode')) {
    // We do not store the data in case the zlib mode is deflate.
    // This should be rarely happening.
    if (zlib_get_coding_type() == 'deflate') {
      $cache = FALSE;
    }
    else if (zlib_get_coding_type() == FALSE) {
      $buffer = gzencode($buffer, 9, FORCE_GZIP);
    }
  }
  
  // Save to cache
  cache_set($key, $buffer, 'cache_page', variable_get('authcache_lifetime', AUTHCACHE_LIFETIME), drupal_get_headers());

}

/**
 * Helper function: Invoke hook_$hook() in all modules, merge recursively
 */
function _authcache_invoke_hook($hook, &$var) {
  foreach (module_implements($hook) as $name) {
    $function = "{$name}_{$hook}";
    $result = $function();
    if (isset($result) && is_array($result)) {
      $var = array_merge_recursive($var, $result);
    }
  }
}


/**
 * Determines the MIME content type of the current page response based on
 * the currently set Content-Type HTTP header.
 *
 * This should normally return the string 'text/html' unless another module
 * has overridden the content type.
 */
function _authcache_get_content_type($default = NULL) {
  static $regex = '!^Content-Type:\s*([\w\d\/\-]+)!i';
  return _authcache_get_http_header($regex, $default);
}

/**
 * Determines the HTTP response code that the current page request will be
 * returning by examining the HTTP headers that have been output so far.
 */
function _authcache_get_http_status($default = 200) {
  static $regex = '!^HTTP/1.1\s+(\d+)!';
  return (int)_authcache_get_http_header($regex, $default);
}

function _authcache_get_http_header($regex, $default = NULL) {
  // The last header is the one that counts:
  $headers = preg_grep($regex, explode("\n", drupal_set_header()));
  if (!empty($headers) && preg_match($regex, array_pop($headers), $matches)) {
    return $matches[1]; // found it
  }
  return $default; // no such luck
}

